from functools import wraps

from .utils import smart_decorator
from .tree import Tree
from .exceptions import VisitError, GrammarError

###{standalone
from inspect import getmembers, getmro

class Discard(Exception):
    pass

# Transformers

class Transformer:
    """Visits the tree recursively, starting with the leaves and finally the root (bottom-up)

    Calls its methods (provided by user via inheritance) according to tree.data
    The returned value replaces the old one in the structure.

    Can be used to implement map or reduce.
    """

    def _call_userfunc(self, tree, new_children=None):
        # Assumes tree is already transformed
        children = new_children if new_children is not None else tree.children
        try:
            f = getattr(self, tree.data)
        except AttributeError:
            return self.__default__(tree.data, children, tree.meta)
        else:
            try:
                if getattr(f, 'meta', False):
                    return f(children, tree.meta)
                elif getattr(f, 'inline', False):
                    return f(*children)
                elif getattr(f, 'whole_tree', False):
                    if new_children is not None:
                        raise NotImplementedError("Doesn't work with the base Transformer class")
                    return f(tree)
                else:
                    return f(children)
            except (GrammarError, Discard):
                raise
            except Exception as e:
                raise VisitError(tree, e)

    def _transform_children(self, children):
        for c in children:
            try:
                yield self._transform_tree(c) if isinstance(c, Tree) else c
            except Discard:
                pass

    def _transform_tree(self, tree):
        children = list(self._transform_children(tree.children))
        return self._call_userfunc(tree, children)

    def transform(self, tree):
        return self._transform_tree(tree)

    def __mul__(self, other):
        return TransformerChain(self, other)

    def __default__(self, data, children, meta):
        "Default operation on tree (for override)"
        return Tree(data, children, meta)

    @classmethod
    def _apply_decorator(cls, decorator, **kwargs):
        mro = getmro(cls)
        assert mro[0] is cls
        libmembers = {name for _cls in mro[1:] for name, _ in getmembers(_cls)}
        for name, value in getmembers(cls):
            if name.startswith('_') or name in libmembers:
                continue

            # Skip if v_args already applied (at the function level)
            if hasattr(cls.__dict__[name], 'vargs_applied'):
                continue

            static = isinstance(cls.__dict__[name], (staticmethod, classmethod))
            setattr(cls, name, decorator(value, static=static, **kwargs))
        return cls


class InlineTransformer(Transformer):   # XXX Deprecated
    def _call_userfunc(self, tree, new_children=None):
        # Assumes tree is already transformed
        children = new_children if new_children is not None else tree.children
        try:
            f = getattr(self, tree.data)
        except AttributeError:
            return self.__default__(tree.data, children, tree.meta)
        else:
            return f(*children)


class TransformerChain(object):
    def __init__(self, *transformers):
        self.transformers = transformers

    def transform(self, tree):
        for t in self.transformers:
            tree = t.transform(tree)
        return tree

    def __mul__(self, other):
        return TransformerChain(*self.transformers + (other,))


class Transformer_InPlace(Transformer):
    "Non-recursive. Changes the tree in-place instead of returning new instances"
    def _transform_tree(self, tree):           # Cancel recursion
        return self._call_userfunc(tree)

    def transform(self, tree):
        for subtree in tree.iter_subtrees():
            subtree.children = list(self._transform_children(subtree.children))

        return self._transform_tree(tree)


class Transformer_InPlaceRecursive(Transformer):
    "Recursive. Changes the tree in-place instead of returning new instances"
    def _transform_tree(self, tree):
        tree.children = list(self._transform_children(tree.children))
        return self._call_userfunc(tree)



# Visitors

class VisitorBase:
    def _call_userfunc(self, tree):
        return getattr(self, tree.data, self.__default__)(tree)

    def __default__(self, tree):
        "Default operation on tree (for override)"
        return tree


class Visitor(VisitorBase):
    """Bottom-up visitor, non-recursive

    Visits the tree, starting with the leaves and finally the root (bottom-up)
    Calls its methods (provided by user via inheritance) according to tree.data
    """


    def visit(self, tree):
        for subtree in tree.iter_subtrees():
            self._call_userfunc(subtree)
        return tree

class Visitor_Recursive(VisitorBase):
    """Bottom-up visitor, recursive

    Visits the tree, starting with the leaves and finally the root (bottom-up)
    Calls its methods (provided by user via inheritance) according to tree.data
    """

    def visit(self, tree):
        for child in tree.children:
            if isinstance(child, Tree):
                self.visit(child)

        f = getattr(self, tree.data, self.__default__)
        f(tree)
        return tree



def visit_children_decor(func):
    "See Interpreter"
    @wraps(func)
    def inner(cls, tree):
        values = cls.visit_children(tree)
        return func(cls, values)
    return inner


class Interpreter:
    """Top-down visitor, recursive

    Visits the tree, starting with the root and finally the leaves (top-down)
    Calls its methods (provided by user via inheritance) according to tree.data

    Unlike Transformer and Visitor, the Interpreter doesn't automatically visit its sub-branches.
    The user has to explicitly call visit_children, or use the @visit_children_decor
    """
    def visit(self, tree):
        return getattr(self, tree.data)(tree)

    def visit_children(self, tree):
        return [self.visit(child) if isinstance(child, Tree) else child
                for child in tree.children]

    def __getattr__(self, name):
        return self.__default__

    def __default__(self, tree):
        return self.visit_children(tree)




# Decorators

def _apply_decorator(obj, decorator, **kwargs):
    try:
        _apply = obj._apply_decorator
    except AttributeError:
        return decorator(obj, **kwargs)
    else:
        return _apply(decorator, **kwargs)



def _inline_args__func(func):
    @wraps(func)
    def create_decorator(_f, with_self):
        if with_self:
            def f(self, children):
                return _f(self, *children)
        else:
            def f(self, children):
                return _f(*children)
        return f

    return smart_decorator(func, create_decorator)


def inline_args(obj):   # XXX Deprecated
    return _apply_decorator(obj, _inline_args__func)



def _visitor_args_func_dec(func, inline=False, meta=False, whole_tree=False, static=False):
    assert [whole_tree, meta, inline].count(True) <= 1
    def create_decorator(_f, with_self):
        if with_self:
            def f(self, *args, **kwargs):
                return _f(self, *args, **kwargs)
        else:
            def f(self, *args, **kwargs):
                return _f(*args, **kwargs)
        return f

    if static:
        f = wraps(func)(create_decorator(func, False))
    else:
        f = smart_decorator(func, create_decorator)
    f.vargs_applied = True
    f.inline = inline
    f.meta = meta
    f.whole_tree = whole_tree
    return f

def v_args(inline=False, meta=False, tree=False):
    "A convenience decorator factory, for modifying the behavior of user-supplied visitor methods"
    if [tree, meta, inline].count(True) > 1:
        raise ValueError("Visitor functions can either accept tree, or meta, or be inlined. These cannot be combined.")
    def _visitor_args_dec(obj):
        return _apply_decorator(obj, _visitor_args_func_dec, inline=inline, meta=meta, whole_tree=tree)
    return _visitor_args_dec


###}



# -*- coding: utf-8 -*-
from gherkan.containers.NLPModelManager import NLPModelManager
from gherkan.containers.signal_batch import SignalBatch
from gherkan.containers.StatementTreeNode import StatementTreeBinaryOperatorNode, \
    StatementTreeNode, StatementTreeOperandNode, StatementTreeMergedNode
from gherkan.containers.StatementTree import StatementTree
from gherkan.processing.TreeProcessor import TreeProcessor

import gherkan.utils.constants as c
import gherkan.utils.gherkin_keywords as g
from gherkan.utils import logging_types

import logging

class SignalFileWriter():
    """ TODO docs """

    def __init__(self, nmm : NLPModelManager):
        self.nlBatch = None
        self.outLines = []
        self.language = None
        self.nmm = nmm
        self.feature = ""

    def encode(self, nlBatch: SignalBatch, no_processing: bool = False):
        self.nlBatch = nlBatch
        self.language = nlBatch.language
        self.no_processing = no_processing
        self.feature = nlBatch.name

        self.outLines.append("# language: {}\n\n".format(self.language))
        self.outLines.append("{}: {}\n".format(g.get_kw(self.language, "Feature"), nlBatch.name))
        self.outLines.append("  {}\n".format(nlBatch.desc))

        if nlBatch.context:
            self.outLines.append("{}:\n".format(g.get_kw(self.language, "Background")))
            self.outLines.append("  {} {}\n\n".format(g.get_kw(self.language, "Given"), self.tree_to_str(nlBatch.context)))

        for scenario in nlBatch.scenarios:
            self.outLines.append("{}: {}\n".format(g.get_kw(self.language, "Scenario"), scenario.name))

            for tree in scenario.givenStatements:
                self.outLines.append("  {} {}\n".format(g.get_kw(self.language, "Given"), self.tree_to_str(tree)))
            for tree in scenario.whenStatements:
                self.outLines.append("  {} {}\n".format(g.get_kw(self.language, "When"), self.tree_to_str(tree)))
            for tree in scenario.thenStatements:
                self.outLines.append("  {} {}\n".format(g.get_kw(self.language, "Then"), self.tree_to_str(tree)))


    def write(self, outputFilePath: str):
        with open(outputFilePath, "w", encoding="utf-8") as out:
            out.writelines(self.outLines)

    def tree_to_str(self, tree: StatementTree):
        if not self.no_processing:
            tp = TreeProcessor(self.language, self.nmm, self.feature)

            # loads yaml file with templates
            tp.load_templ_dic('utils/templates_dic.yaml')
            tree.root = tp.process_tree(tree.root, TreeProcessor.Direction.NL_TO_SIGNAL)

        return self.node_to_str(tree.root)

    def node_to_str(self, node: StatementTreeNode):
        if type(node) == StatementTreeBinaryOperatorNode:
            return self.operator_to_str(node)
        elif type(node) == StatementTreeMergedNode:
            return self.merged_operator_to_str(node)
        elif type(node) == StatementTreeOperandNode:
            return self.operand_to_str(node)

    def merged_operator_to_str(self, node: StatementTreeMergedNode):
        return " && ".join([self.operand_to_str(subnode) for subnode in node.subnodes])

    def operator_to_str(self, node: StatementTreeBinaryOperatorNode):
        if node.kind == c.AND:
            operator = "&&"
        elif node.kind == c.OR:
            operator = "||"
        else:
            raise NotImplementedError("Unrecognized operator {}".format(node.kind))

        return ("({}) {} ({})".format(
            self.node_to_str(node.lchild),
            operator,
            self.node_to_str(node.rchild)))

    def operand_to_str(self, node: StatementTreeOperandNode):
        kind = node.kind

        if kind == c.EQUALITY:
            return "{} {}= {}".format(
                node.data.variable,
                "!" if node.negated else "=",
                node.data.value
            )
        elif kind == c.BOOL:
            return "{}{}".format(
                # XOR in case boolean is negated by the 'negated' attribute
                "" if (node.data.value ^ node.negated) else "! ",
                node.data.variable,
            )
        elif kind in [c.EDGE,
                      c.FORCE,
                      c.UNFORCE]:
            return "{}{}({}, {})".format(
                "! " if node.negated else "",
                node.kind.lower(),
                node.data.variable,
                node.data.value
            )
        else:
            logging.warning("Node kind {} not recognized!", kind, extra={"type": logging_types.W_OPERAND_UNKNOWN, "phrase": ""})



from gherkan.decoder.NLParser import NLParser
from gherkan.encoder.SignalFileWriter import SignalFileWriter
import importlib
import os
from gherkan.containers.NLPModelManager import NLPModelManager

import gherkan.utils.constants as c

# Path to the text file with the signal batch
input_path = os.path.join(c.DATA_DIR, 'output')
output_path = os.path.join(c.DATA_DIR, 'output')
input_file = os.path.join(input_path, 'case_NLParser.feature')
output_file = os.path.join(output_path, 'case_NLParser_signals.feature')

nmm = NLPModelManager()
nmm.load_all_models()

nlFile = NLParser()
# Parse the entire file
nlBatch = nlFile.parseFile(input_file)

signalFileWriter = SignalFileWriter(nmm)
signalFileWriter.encode(nlBatch)
signalFileWriter.write(output_file)

with open(output_file, "rt", encoding="utf-8") as f:
    text = f.read()
    print(text)



import sys
from collections import deque

Py36 = (sys.version_info[:2] >= (3, 6))

class fzset(frozenset):
    def __repr__(self):
        return '{%s}' % ', '.join(map(repr, self))


def classify_bool(seq, pred):
    true_elems = []
    false_elems = []

    for elem in seq:
        if pred(elem):
            true_elems.append(elem)
        else:
            false_elems.append(elem)

    return true_elems, false_elems

def classify(seq, key=None, value=None):
    d = {}
    for item in seq:
        k = key(item) if (key is not None) else item
        v = value(item) if (value is not None) else item
        if k in d:
            d[k].append(v)
        else:
            d[k] = [v]
    return d

def bfs(initial, expand):
    open_q = deque(list(initial))
    visited = set(open_q)
    while open_q:
        node = open_q.popleft()
        yield node
        for next_node in expand(node):
            if next_node not in visited:
                visited.add(next_node)
                open_q.append(next_node)




###{standalone
try:
    STRING_TYPE = basestring
except NameError:   # Python 3
    STRING_TYPE = str


import types
from functools import wraps, partial
from contextlib import contextmanager

Str = type(u'')
try:
    classtype = types.ClassType # Python2
except AttributeError:
    classtype = type    # Python3

def smart_decorator(f, create_decorator):
    if isinstance(f, types.FunctionType):
        return wraps(f)(create_decorator(f, True))

    elif isinstance(f, (classtype, type, types.BuiltinFunctionType)):
        return wraps(f)(create_decorator(f, False))

    elif isinstance(f, types.MethodType):
        return wraps(f)(create_decorator(f.__func__, True))

    elif isinstance(f, partial):
        # wraps does not work for partials in 2.7: https://bugs.python.org/issue3445
        return create_decorator(f.__func__, True)

    else:
        return create_decorator(f.__func__.__call__, True)

def dedup_list(l):
    """Given a list (l) will removing duplicates from the list,
       preserving the original order of the list. Assumes that
       the list entrie are hashable."""
    dedup = set()
    return [ x for x in l if not (x in dedup or dedup.add(x))]

###}


try:
    from contextlib import suppress     # Python 3
except ImportError:
    @contextmanager
    def suppress(*excs):
        '''Catch and dismiss the provided exception

        >>> x = 'hello'
        >>> with suppress(IndexError):
        ...     x = x[10]
        >>> x
        'hello'
        '''
        try:
            yield
        except excs:
            pass




try:
    compare = cmp
except NameError:
    def compare(a, b):
        if a == b:
            return 0
        elif a > b:
            return 1
        return -1


import sre_parse
import sre_constants
def get_regexp_width(regexp):
    try:
        return sre_parse.parse(regexp).getwidth()
    except sre_constants.error:
        raise ValueError(regexp)



import requests
import unittest

class TestNLText(unittest.TestCase):

    def setUp(self):
        self.request = {
            "feature": "Montrac",
            "feature_desc": " Montrac",
            "background": "Given line is on",
            "text_raw": "scenario given line is on",
            "language": "en"
        }
        self.url = f"http://{self.__class__.host}:{self.__class__.port}/{{}}"

        resetResponse = requests.get(self.url.format("reset"))
        self.assertDictContainsSubset({"OK": True}, resetResponse.json())

    # @unittest.skip("TODO: skipped for debugging")
    def test_post_nltext(self):
        correctResponse = {
            "info": "Processing done",
            "lines": [],
            "error_lines": [],
            "error_hints": [],
            "errors": False
        }

        response = requests.post(self.url.format(
            "nltext"), data=self.request).json()
        self.assertDictContainsSubset(correctResponse, response)

    # @unittest.skip("TODO: skipped for debugging")
    def test_post_and_get_signal(self):
        correctResponse = """# language: en
            Feature: Montrac
              Montrac
            Background:
              Given lineon
            
            Scenario: given line is on
              Given lineon
            """

        response = requests.post(self.url.format("nltext"), data=self.request)
        try:
            response = response.json()
        except Exception as exception:
            self.fail("An exception was raised while trying to convert response to JSON. "
                      "This is most likely due to the fact that the response was of different type.\n"
                      "The exception: {}\nResponse text (if possible):\n{}".format(exception, response.text))
        else:
            self.assertDictContainsSubset({"errors": False}, response)

            response = requests.get(self.url.format("signals"))
            self.compare_multiline(response.content.decode(), correctResponse)

    def compare_multiline(self, testee, teststr):
        if not isinstance(testee, str) or not isinstance(teststr, str):
            self.assertFalse("No strings given to compare!")
            return

        self.assertEqual(testee.lower().replace("\n", " ").split(),
                         teststr.lower().replace("\n", " ").split())


        
if __name__ == '__main__':
    unittest.TestCase.host = "localhost"
    unittest.TestCase.port = 5000
    unittest.main()


# -*- coding: utf-8 -*-

import yaml


class Translator():

    def __init__(self, dict_path):
        try:
            dictFile = open(dict_path, mode="r", encoding="utf-8")
        except IOError as e:
            print("There was an error reading the dictionary file: {}".format(e))
            return
        self.dictionary = {}
        self.lookup = {}    # inverted dictionary
        self.conseq = {}    # list of words after which other words can appear
        self.alternativeWords = []
        self.__parseDict(yaml.full_load(dictFile))

    def __parseDict(self, d):
        if "listMode" in d:
            self.listMode = d["listMode"]
        else:
            self.listMode = "hmps"
        dictionary = d["dictionary"]

        for (word, alts) in dictionary.items():
            alternatives, hmps, syms = [], [], []
            if isinstance(alts, dict):
                keys = list(alts.keys())
                if "hmps" in keys:
                    hmps = alts["hmps"]
                    keys.remove("hmps")
                if "syms" in keys:
                    syms = alts["syms"]
                    keys.remove("syms")

                # prevent from misspelling hmps or syms
                if len(keys) > 0:
                    raise Exception("Warning, unknown word variation(s) in the dictionary: {}".format(str(keys)))
            elif isinstance(alts, list):
                if self.listMode == "hmps":
                    hmps = alts
                elif self.listMode == "syms":
                    syms = alts
            else:
                raise Exception("Error: unsupported form ({}) in which alternatives to the word {} have been presented.".format(type(alts), word))

            alternatives = hmps + syms

            if word not in self.dictionary:
                self.dictionary[word] = {}
                self.dictionary[word]["hmps"] = hmps
                self.dictionary[word]["syms"] = syms
                self.dictionary[word]["alts"] = alternatives
            else:
                self.dictionary[word]["hmps"] += hmps
                self.dictionary[word]["syms"] += syms
                self.dictionary[word]["alts"] += alternatives

            for alt in alternatives:
                if " " in alt:
                    altsplit = alt.split(" ")
                    self.__addConsecutiveAlts(altsplit, self.conseq)
                if alt in self.lookup:
                    raise Exception("Error: ambiguous alternative. The word {} was provided as an alternative to {} but it was already assigned as an alternative to {}!".format(alt, word, self.lookup[alt]))
                self.lookup[alt] = word

            self.alternativeWords += self.lookup.keys()

    def __addConsecutiveAlts(self, altRemainder, conseq):
        """ Recursion warning - this function is awful and attempt to understand it requires high amounts of caffeine and carbohydrates
        But essentially, it just traverses a nested dictionary and tries to put words in the right place to create
        chains of words that will be used when parsing phrases in the sentences.
        """
        alt = altRemainder[0]
        nextAlt = altRemainder[1]
        lastPair = len(altRemainder) == 2
        if alt in conseq:
            wcsq = conseq[alt]
            if isinstance(wcsq, str):
                if wcsq == nextAlt:
                    if lastPair:
                        pass  # this should not happen as it means the same phrase was added twice
                    else:
                        conseq[alt] = {wcsq: {".": None}}
                else:
                    if lastPair:
                        conseq[alt] = {wcsq: ".", nextAlt: "."}
                    else:
                        conseq[alt] = {wcsq: ".", nextAlt: {}}
            else:  # wcsq is a dictionary
                if lastPair:
                    if nextAlt in wcsq:
                        if isinstance(wcsq[nextAlt], str):
                            wcsq[nextAlt] = {wcsq[nextAlt]: ".", ".": None}
                        else:
                            wcsq[nextAlt]["."] = None
                    else:
                        wcsq[nextAlt] = "."
        else:  # alt is not in conseq
            if lastPair:
                conseq[alt] = nextAlt
            else:
                conseq[alt] = {}

        if not lastPair:
            self.__addConsecutiveAlts(altRemainder[1:], conseq[alt])

    def __checkConsecutiveWords(self, sentRemainder, conseq):
        """ Recursion warning - this function is awful and attempt to understand it requires high amounts of caffeine and carbohydrates
        But all it does is it tries to traverse the nested dictionary created by the "__addConsecutiveAlts" function
        to properly parse the sentence and attempt to find predefined phrases.
        """
        w = sentRemainder[0]
        if conseq is None or not sentRemainder:
            return []  # if there are no words remaining in the sentence - this should only happen in case of error
        elif isinstance(conseq, str):
            if w == conseq:
                return [conseq]  # if the consecutives is just a single word, then return that word and end
            elif '.' == conseq:
                return True  # current word is not in consequtives but the previous word can be terminal
            else:

                return []
        else:  # else, there are some consecutives to be processed
            followers = []  # words following this word in the phrase
            if w in conseq:
                if len(sentRemainder) > 1:
                    followers = self.__checkConsecutiveWords(sentRemainder[1:], conseq[w])
                elif '.' in conseq[w]:  # dot in the consecutives dictionary means this word can be terminal
                    return [w]  # valid last word in the phrase (not necessarily in the sentence)
                else:
                    return []
            elif '.' in conseq:
                return True  # current word is not in consequtives but the previous word can be terminal
            else:
                return []

            if not followers:
                return []
            elif isinstance(followers, list):
                return [w] + followers
            else:
                return [w]  # the word has no followers in the sentence but was recognized as a valid terminal word

    def translate(self, words):
        """ words - a single word or a list of words
            returns a list of translated words
        """
        if isinstance(words, str) and " " in words:
            return self.translate(words.split(" "))
        elif isinstance(words, list):
            sentence = []
            n = len(words)
            i = 0
            while i < n:
                w = words[i]
                if w in self.conseq and i < n - 1:  # checks for alternative multiwords phrases (instead of a single word)
                    phrase = self.__checkConsecutiveWords(words[i+1:], self.conseq[w])
                    if len(phrase) > 0:  # length check to prevent boolean True from sneaking in, which should only happen in case of an error
                        phrase = [w] + phrase
                        i += len(phrase) - 1
                        w = ' '.join(phrase)
                        sentence.append(self.lookup[w])
                else:
                    sentence.append(self.translate(w))
                i += 1
            return sentence
        else:
            if words in self.alternativeWords:
                return self.lookup[words]
            else:
                return words



import os

GHERKAN_ROOT_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir)
DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir, os.pardir, "data")
SUBJ_OF_INTEREST = "robot"

AND = "AND"
OR = "OR"
NOT = "NOT"
EQUALITY = "EQUALITY"
INEQUALITY = "INEQUALITY"
BOOL = "BOOL"
EDGE = "EDGE"
FORCE = "FORCE"
UNFORCE = "UNFORCE"

LBRACKET = "LBRACKET"
RBRACKET = "RBRACKET"
SKIP = "SKIP"
MISMATCH =  "MISMATCH"

LANG_CZ = "cs"
LANG_EN = "en"


#
# This example shows how to write a basic calculator with variables.
#

from lark import Lark
from lark import tree as lark_tree

statement_grammar = """
    ?start: expression

    ?expression: statement
        | "(" expression ")"
        | expression "&&" expression        -> AND
        | expression "||" expression        -> OR
        | "!" expression                    -> NOT
    
    ?statement: function
        | VARIABLE "==" VALUE               -> EQUALITY
        | VARIABLE "!=" VALUE               -> INEQUALITY
        | VARIABLE
        
    ?function: function_name "(" VARIABLE "," VALUE ")"
    
    function_name: "edge" | "force" | "unforce"
    
    VARIABLE: NAME
    VALUE: NUMBER

    %import common.CNAME -> NAME
    %import common.NUMBER
    %import common.WS_INLINE
    
    %ignore WS_INLINE
"""


parser = Lark(statement_grammar, parser='earley', ambiguity='explicit')

def test():

    statement = "LinkaOn == 1 || (robotR1ProgramNumber == 1 && robotR2ProgramNumber !=  2 && edge(shuttleAtXY, 1))"

    tree = parser.parse(statement)
    print(tree.pretty())

    lark_tree.pydot__tree_to_png(tree, "statement.png")


if __name__ == '__main__':
    test()
    # main()



import requests
import unittest

import gherkan.utils.gherkin_keywords as g
import gherkan.utils.constants as c

class TestNegatedSignalStatements(unittest.TestCase):

    def setUp(self):
        self.url = f"http://{self.__class__.host}:{self.__class__.port}/{{}}"

        resetResponse = requests.get(self.url.format("reset"))
        self.assertDictContainsSubset({"OK": True}, resetResponse.json())



    def test_negate_signal_file(self):
        lang = "en"
        signalFile = f"""
            # language: {lang}

            Feature: StatementTest
              StatementTest dummy description
            Background:
              Given lineOn

            Scenario: dummy description
              Given ((shuttleYDestination == XZZ) || stationXZXFree) && robotR1ProgramNumber == 2
            """

        correct_response = f"""# language: {lang}

            Feature: StatementTest
              StatementTest dummy description
            Background:
              Given ! lineOn

            Scenario: dummy description
              Given ((shuttleYDestination != XZZ) && (! stationXZXFree)) || (robotR1ProgramNumber != 2)
            """

        request = {'language': lang, 'scenarios': signalFile}

        response = requests.post(self.url.format("signals"), data=request)
        try:
            response = response.json()
        except Exception as exception:
            self.fail(
                "An exception was raised while trying to convert response to JSON. "
                "This is most likely due to the fact that the response was of different type.\n"
                "The exception: {}\nResponse text (if possible):\n{}".format(
                    exception, response.text))
        else:
            self.assertDictContainsSubset({"OK": True}, response)

            response = requests.get(self.url.format("signals_negated"))
            content = response.content.decode()

            content_list = [s.strip().lower() for s in content.split("\n")]
            correct_response_list = [s.strip().lower() for s in correct_response.split("\n")]

            # case correct_response_list
            self.assertEqual(content_list, correct_response_list)



if __name__ == '__main__':
    unittest.TestCase.host = "localhost"
    unittest.TestCase.port = 5000
    unittest.main()



import os
import spacy
import majka

from ufal.morphodita import Tagger
from gherkan.utils import constants as c


class NLPModelManager:

    def __init__(self):
        self.nlp_model_en = None
        self.nlp_model_cs = None
        self.morph_lwt = None
        self.morph_wlt = None

        self.nlp_model_cs_path = os.path.join(c.GHERKAN_ROOT_DIR, "morphodita", "czech-morfflex-pdt-161115",
                                   "czech-morfflex-pdt-161115.tagger")

        self.morph_lwt_path = os.path.join(c.GHERKAN_ROOT_DIR, "utils", "majka.l-wt")
        self.morph_wlt_path = os.path.join(c.GHERKAN_ROOT_DIR, "utils", "majka.w-lt")


    def load_all_models(self):
        self.nlp_model_en = spacy.load('en_core_web_sm', disable=['tagger', 'parser', 'ner'])
        self.nlp_model_cs = Tagger.load(self.nlp_model_cs_path)

        self.morph_lwt = majka.Majka(self.morph_lwt_path)
        self.morph_wlt = majka.Majka(self.morph_wlt_path)


from .actions import Actions
from .audio import Audio
from .nl_scenario import NLScenario
from .nl_text import NLText
from .signal_map import SignalMap
from .signals import Signals
from .signals import NegatedSignals
from .handler import Handler


from gherkan.decoder.SignalParser import SignalParser

sp = SignalParser()
st = sp.parseStatement("part2OnTable == 1 && (robotN1ProgramNumber == 0 || robotN1ProgramNumber == 1 || robotN1ProgramEnded) && edge(robotN3ProgramEnded, 1)")

print(st)


from setuptools import setup
from setuptools.command.install import install as _install
from setuptools.command.develop import develop as _develop
from setuptools.command.build_py import build_py as _build_py
from setuptools import find_packages
import re
import sys
from subprocess import run

with open('requirements.txt') as f:
    requirements = [l.strip() for l in f]

with open('dependency_links.txt') as f:
    dependency_links = [l.strip() for l in f]

# filter non-standard requirements
reqexp = re.compile(r"[^\w><=\.\s-]")
nonstandard = list(filter(reqexp.search, requirements))
requirements = list(filter(lambda w: not(reqexp.search(w)), requirements))

if nonstandard:
    if sys.argv[1] != "clean":
        print("Non-standard requirements found. These will have to installed manually. The non-standard requirements are:")
        print(nonstandard)

with open("README.md", "r") as f:
    long_description = f.read()

with open("LICENSE.txt", "r") as f:
    license_text = f.read()

versionRegex = re.compile("\d+[.]\d+[.]\d+[^\s]*")
with open("CHANGELOG.md", "r") as f:
    for line in f:
        versionMatch = versionRegex.match(line)
        if versionMatch:
            version = versionMatch.group()
            break

class BuildWrapper(_build_py):

    def run(self):
        print(f"Building version {version}")
        init_file = "gherkan/__init__.py"
        fileText = [
            '# This file was generated by the setup.py and will be rewritten at the next build\n',
            f'__version__ = "{version}"\n',
            '__name__ = "gherkan"'
        ]
        with open(init_file, "w") as f:
            f.writelines(fileText)
        _build_py.run(self)


class InstallWrapper(_install):

    def run(self):
        print(">>>>>>>>>>>>>>>>>>>>>>>>>>>")
        _install.run(self)
        print("<<<<<<<<<<<<<<<<<<<<<<<<<<<")
        self.execute(self._post_install, (),
                     msg="Running post install task")

    def _post_install(self):
        # TODO: check spacy file version SM/MD/LG
        print("installing neural coref")
        result = run(["pip", "install", "https://github.com/huggingface/neuralcoref-models/releases/download/en_coref_md-3.0.0/en_coref_md-3.0.0.tar.gz"])
        if result.returncode == 0:
            print(result.stdout)
        else:
            if result.stderr:
                print('Preprocess failed: ')
                print(result.stderr)
            print("Downloading neuralcoref failed, please run:\n'pip install https://github.com/huggingface/neuralcoref-models/releases/download/en_coref_md-3.0.0/en_coref_md-3.0.0.tar.gz'")
       
        print("Downloading en_core_web_sm")
        result = run(["python", "-m", "spacy", "link", "en_core_web_sm", "en", "--force"])  #, capture_output=True)
        if result.returncode == 0:
            print(result.stdout)
        else:
            if result.stderr:
                print('Preprocess failed: ')
                print(result.stderr)
            print("Downloading en_core_web_sm failed, please run:\n'python -m spacy download en_core_web_sm'")
        print("creating morphodita folder and downloading models")
        result = run(["sh", "install_morphodita.sh"])
        if result.returncode == 0:
            print(result.stdout)
        else:
            if result.stderr:
                print('Preprocess failed: ')
                print(result.stderr)
            print("Downloading morphodita failed, please make folder for morphodita and unzip there the remote model 'curl --remote-name-all","https://lindat.mff.cuni.cz/repository/xmlui/bitstream/handle/11234/1-1836{/czech-morfflex-pdt-161115.zip}'")
        


class DevelopWrapper(_develop):

    def run(self):
        print(">>>>>>>>>>>>>>>>>>>>>>>>>>>")
        _develop.run(self)
        print("<<<<<<<<<<<<<<<<<<<<<<<<<<<")
        self.execute(self._post_install, (),
                     msg="Running post install task")

    def _post_install(self):
        # TODO: check spacy file version SM/MD/LG
        print("installing neural coref")
        result = run(["pip", "install", "https://github.com/huggingface/neuralcoref-models/releases/download/en_coref_md-3.0.0/en_coref_md-3.0.0.tar.gz"])
        if result.returncode == 0:
            print(result.stdout)
        else:
            if result.stderr:
                print('Preprocess failed: ')
                print(result.stderr)
            print("Downloading neuralcoref failed, please run:\n'pip install https://github.com/huggingface/neuralcoref-models/releases/download/en_coref_md-3.0.0/en_coref_md-3.0.0.tar.gz'")       
        print("Downloading en_core_web_sm")
        result = run(["python", "-m", "spacy", "link", "en_core_web_sm", "en", "--force"])#, capture_output=True)
        if result.returncode == 0:
            print(result.stdout)
        else:
            if result.stderr:
                print('Preprocess failed: ')
                print(result.stderr)
            print("Downloading en_core_web_sm failed, please run:\n'python -m spacy download en_core_web_sm'")
        print("creating morphodita folder and downloading models")
        result = run(["sh", "install_morphodita.sh"])
        if result.returncode == 0:
            print(result.stdout)
        else:
            if result.stderr:
                print('Preprocess failed: ')
                print(result.stderr)
            print("Downloading morphodita failed, please make folder for morphodita and unzip there the remote model 'curl --remote-name-all","https://lindat.mff.cuni.cz/repository/xmlui/bitstream/handle/11234/1-1836{/czech-morfflex-pdt-161115.zip}'")
        

setup(
    name='gherkan',
    version=version,
    # use_scm_version=True,
    description='NL to Gherkin format translation tool',
    long_description=long_description,
    author='Imitation Learning Group CIIRC CVUT',
    maintainer_email='radoslav.skoviera@cvut.cz',
    url='',
    download_url='',
    license=license_text,
    install_requires=requirements,
    dependency_links=dependency_links,
    include_package_data=True,
    # setup_requires=['setuptools_scm'],
    packages=find_packages(exclude=["test", "data"]),
    cmdclass={
        'build_py': BuildWrapper,
        'install': InstallWrapper,
        'develop': DevelopWrapper
        },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Development Status :: 3 - Alpha",
        "Framework :: Flask",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Natural Language :: Czech",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Human Machine Interfaces"
    ]
)



# -*- coding: utf-8 -*-
import re

import pkg_resources
import yaml
import logging

from gherkan.containers.NLPModelManager import NLPModelManager
from gherkan.utils import logging_types

from gherkan.containers.SignalPhrase import SignalPhrase
from gherkan.containers.StatementTreeNode import StatementTreeBinaryOperatorNode, StatementTreeOperandNode, \
    StatementTreeNode, StatementTreeMergedNode

from gherkan.encoder.NL2Temp import NLTempMatcher

from enum import Enum
from pprint import pprint as pp
from gherkan.processing.Temp2NL import FillActions
from gherkan.utils import constants as c


class TreeProcessor:
    class Direction(Enum):
        SIGNAL_TO_NL = 0
        NL_TO_SIGNAL = 1

    def __init__(self, language : str, nmm : NLPModelManager, feature: str):
        self.templ_dic = None
        self.language = language
        self.nmm = nmm

        self.ntm = NLTempMatcher(lang=self.language, nmm=self.nmm, feature=feature)


    def process_tree(self, node: StatementTreeNode, dir: Direction):
        if dir not in [TreeProcessor.Direction.SIGNAL_TO_NL,
                       TreeProcessor.Direction.NL_TO_SIGNAL]:

            err_msg = "Direction {} unknown".format(dir)
            logging.error(err_msg)

        self.dir = dir


        return self.process_tree_recursive(node)


    def process_tree_recursive(self, node: StatementTreeNode):
        if type(node) is StatementTreeBinaryOperatorNode:

            if self.dir == TreeProcessor.Direction.SIGNAL_TO_NL and node.kind == c.AND:
                # try to match compound template which can be rooted in the AND node
                res = self.match_compound_signal_templates(node)

                # in case the compound template has been matched, replace the AND node
                # at the parent by the new merged node
                if res:
                    template, matches, merged_node = res
                    self.process_operand(merged_node, matches, template)
                    return merged_node

            node.lchild = self.process_tree_recursive(node.lchild)
            node.rchild = self.process_tree_recursive(node.rchild)

        elif type(node) is StatementTreeOperandNode:

            if self.dir == TreeProcessor.Direction.NL_TO_SIGNAL:
                # normalize NL phrase before template matching
                res = self.ntm.get_NLPhrase(node.data.variableNL)
                node.data.variableNL = res["tempToNL"]

                self.add_node_modifiers(node, res)

                match, template = self.match_simple_templates(node)

                if match is None:
                    logging.warning("No match found for phrase '{}'.", node.data.variableNL,
                                    extra={"type": logging_types.W_TEMPLATE_NOT_FOUND, "phrase": node.data.variableNL})
                    return

                node = self.process_operand(node, [match], template)

            elif self.dir == TreeProcessor.Direction.SIGNAL_TO_NL:
                match, template = self.match_simple_templates(node)

                if match is None:
                    logging.warning("No match found for phrase '{}'.", node.data.variableNL,
                                    extra={"type": logging_types.W_TEMPLATE_NOT_FOUND, "phrase": node.data.string})
                    return

                node = self.process_operand(node, [match], template)

        else:
            err_msg = "Unrecognized type of node: {}".format(type(node))
            logging.error(err_msg)

        # returned node should be reassigned in case the node changed
        return node



    def add_node_modifiers(self, node, res):
        if res["state"] is not None:
            # TODO this is only a temporary solution: state is appended at the end of the phrase and matched with a template
            # TODO should be handled by NLTempMatcher
            node.data.variableNL += ":" + res["state"]

        if res["value"] is not None:
            node.data.value = res["value"]

        if res["negate"]:
            node.negate()
        elif res["force"]:
            node.kind = c.FORCE
        elif "edge" in res and res["edge"]:  # TODO remove ("edge" in res) -> should be there always
            node.kind = c.EDGE


    def process_operand(self, node: StatementTreeNode, matches: list, template):
        sp = SignalPhrase()
        sp.tempName = template["Name"]
        sp.tempStr = template["SignalTemplate"]
        sp.tempToNL = template["NLTemplateEn"] if self.language == c.LANG_EN else template["NLTemplateCz"]
        sp.inclinations = template["Inclinations"]
        sp.language = self.language
        sp.negate = node.isNegated()
        sp.force = (node.kind == c.FORCE)
        sp.edge = (node.kind == c.EDGE)

        for match in matches:
            for key, val in match.groupdict().items():
                sp.vars[key] = val

        if self.dir == TreeProcessor.Direction.SIGNAL_TO_NL:
            if type(node) == StatementTreeOperandNode:
                sp.values[node.data.valueId] = node.data.value

            elif type(node) == StatementTreeMergedNode:
                for subnode in node.subnodes:
                    sp.values[subnode.data.valueId] = subnode.data.value

            self.translate_to_nl(node, sp)
        elif self.dir == TreeProcessor.Direction.NL_TO_SIGNAL:
            node = self.translate_to_signal(node, sp, template)

        return node



    def translate_to_signal(self, node: StatementTreeOperandNode, sp: SignalPhrase, template: dict):
        signal_templates = template["SignalTemplate"]

        if len(signal_templates) > 1:
            mergedNode = StatementTreeMergedNode(parent=node.parent)

            for signal_template in signal_templates:
                subnode = StatementTreeOperandNode(kind=None) # kind is filled by translate_node_to_signal()
                subnode = self.translate_node_to_signal(subnode, sp, signal_template)
                mergedNode.subnodes.append(subnode)

            return mergedNode

        else:
            return self.translate_node_to_signal(node, sp, template["SignalTemplate"][0])


    def translate_node_to_signal(self, node: StatementTreeOperandNode, sp: SignalPhrase, signal_template: dict):
        variable = signal_template["Template"]

        # replace all named patterns by their value
        for group, value in sp.vars.items():
            if value is not None:
                value = self.ntm.translate(value)
                # capitalize the first letter of every value inside a variable
                # note: do not use variable.capitalize(), this would make all other letters lowercase
                value = value[0].upper() + value[1:]
                strToReplace = re.compile(r"\(\?P\<{}\>[^\)]+\)".format(group))
                variable = re.sub(strToReplace, value, variable)

        # the result should be a variable name without any whitespaces
        variable = re.sub(r'\s+', '', variable)

        # make the variable camelcase - lowercase the first letter
        variable = variable[0].lower() + variable[1:]

        node.data.variable = variable

        # in case it's not edge or force, just use the template type
        if node.kind is None:
            node.kind = signal_template["Type"]


        # fill in the value
        valueId = signal_template["ValueId"]

        if node.data.value is None:
            if valueId in sp.vars:
                # should work for every template where value is stated explicitly in the NL phrase
                var = self.ntm.translate(sp.vars[valueId])
                node.data.value = var
            elif node.kind == c.BOOL:
                # default value for bool
                node.data.value = True

                if node.negated:
                    node.data.value = False
                    node.negated = False
            else:
                # TODO should not occur in practice
                # default value for everything else
                node.data.value =  1

        return node


    def translate_to_nl(self, node: StatementTreeNode, signalPhrase: SignalPhrase):
        # call Temp2NL
        fill_action = FillActions(self.nmm)
        fill_action.ParseMainAction(data=signalPhrase)

        if not signalPhrase.niceStr:
            logging.warning("niceStr not received in the response",
                            extra={"type": logging_types.W_GENERAL_ERROR, "phrase": signalPhrase})

        node.data.variableNL = signalPhrase.niceStr


    def match_compound_signal_templates(self, node: StatementTreeBinaryOperatorNode):
        # only for direction TreeProcessor.Direction.SIGNAL_TO_NL
        operands = self.gather_conjunction_operands(node)

        for template in self.templ_dic:
            if not operands or len(template["SignalTemplate"]) != len(operands):
                # number of operands do not equal number of template strings,
                # cannot match the compound template
                continue

            matched_str_idxs = []
            matches = []

            for operand in operands:
                i, match = self.match_signal_template(operand, template)

                if match:
                    matches.append(match)
                    # save the index of the matched template string
                    matched_str_idxs.append(i)

            # check if matched templates strings are unique (none was matched twice)
            # and complete (all were matched)
            if len(set(matched_str_idxs)) == len(operands):
                merged_node = StatementTreeMergedNode()
                merged_node.subnodes = operands

                return template, matches, merged_node

        return None


    def gather_conjunction_operands(self, node):
        # gathers a list of operand nodes in a subtree connected via AND operator nodes
        if node.kind == c.AND:
            left_operands = self.gather_conjunction_operands(node.lchild)
            right_operands = self.gather_conjunction_operands(node.rchild)

            if left_operands and right_operands:
                # concatenate lists from both branches
                return left_operands + right_operands

            # if one of the branches did not succeed, the whole operation failed

        elif type(node) is StatementTreeOperandNode:
            # we encountered an operand, return it as an element of a list
            return [node]

        # signalizes the operation was invalid, subtree is not a conjunction of operands
        return None




    def match_simple_templates(self, node: StatementTreeOperandNode):
        # finds a matching template in the YAML file (works for both directions)
        match = None

        for template in self.templ_dic:
            if self.dir == TreeProcessor.Direction.NL_TO_SIGNAL:
                match = self.match_nl_template(node, template)
            elif self.dir == TreeProcessor.Direction.SIGNAL_TO_NL:

                if len(template["SignalTemplate"]) > 1:
                    # here we match only simple templates
                    continue

                _, match = self.match_signal_template(node, template)

            if match:
                return match, template

        return None, None



    def match_signal_template(self, operand: StatementTreeOperandNode, template: dict):
        templates_str = template["SignalTemplate"]

        # at least one of the SignalTemplates has to match operand variable
        for i, template_str in enumerate(templates_str):
            expToMatch = re.compile(r"^{}$".format(template_str["Template"].strip()))
            match = expToMatch.search(operand.data.variable)

            if match:
                operand.data.valueId = template_str["ValueId"]
                return i, match

        # no match found
        return None, None

    def match_nl_template(self, operand: StatementTreeOperandNode, template: dict):
        templates_str = template["NLTemplateEn"
                            if self.language == c.LANG_EN
                            else "NLTemplateCz"]


        for i, template_str in enumerate(templates_str):
            expToMatch = re.compile(r"^\s*{}\s*$".format(template_str), flags=re.IGNORECASE)
            match = expToMatch.search(operand.data.variableNL)

            if match:
                return match

        return None


    def load_templ_dic(self, file):
        # load templates to be matched
        with pkg_resources.resource_stream("gherkan", file) as stream:
            try:
                self.templ_dic = yaml.full_load(stream)
            except yaml.YAMLError as exc:
                logging.exception(exc)


#
# This example demonstrates relative imports with rule rewrite
# see multiples.lark
#

#
# if b is a number written in binary, and m is either 2 or 3,
# the grammar aims to recognise m:b iif b is a multiple of m
#
# for example, 3:1001 is recognised
# because 9 (0b1001) is a multiple of 3
#

from lark import Lark, UnexpectedInput

parser = Lark.open('multiples.lark', parser='lalr')

def is_in_grammar(data):
    try:
        parser.parse(data)
    except UnexpectedInput:
        return False
    return True

for n_dec in range(100):
    n_bin = bin(n_dec)[2:]
    assert is_in_grammar('2:{}'.format(n_bin)) == (n_dec % 2 == 0)
    assert is_in_grammar('3:{}'.format(n_bin)) == (n_dec % 3 == 0)



from .utils import STRING_TYPE

###{standalone
class LarkError(Exception):
    pass

class GrammarError(LarkError):
    pass

class ParseError(LarkError):
    pass

class LexError(LarkError):
    pass

class UnexpectedInput(LarkError):
    pos_in_stream = None

    def get_context(self, text, span=40):
        pos = self.pos_in_stream
        start = max(pos - span, 0)
        end = pos + span
        before = text[start:pos].rsplit('\n', 1)[-1]
        after = text[pos:end].split('\n', 1)[0]
        return before + after + '\n' + ' ' * len(before) + '^\n'

    def match_examples(self, parse_fn, examples):
        """ Given a parser instance and a dictionary mapping some label with
            some malformed syntax examples, it'll return the label for the
            example that bests matches the current error.
        """
        assert self.state is not None, "Not supported for this exception"

        candidate = None
        for label, example in examples.items():
            assert not isinstance(example, STRING_TYPE)

            for malformed in example:
                try:
                    parse_fn(malformed)
                except UnexpectedInput as ut:
                    if ut.state == self.state:
                        try:
                            if ut.token == self.token:  # Try exact match first
                                return label
                        except AttributeError:
                            pass
                        if not candidate:
                            candidate = label

        return candidate


class UnexpectedCharacters(LexError, UnexpectedInput):
    def __init__(self, seq, lex_pos, line, column, allowed=None, considered_tokens=None, state=None):
        message = "No terminal defined for '%s' at line %d col %d" % (seq[lex_pos], line, column)

        self.line = line
        self.column = column
        self.allowed = allowed
        self.considered_tokens = considered_tokens
        self.pos_in_stream = lex_pos
        self.state = state

        message += '\n\n' + self.get_context(seq)
        if allowed:
            message += '\nExpecting: %s\n' % allowed

        super(UnexpectedCharacters, self).__init__(message)



class UnexpectedToken(ParseError, UnexpectedInput):
    def __init__(self, token, expected, considered_rules=None, state=None):
        self.token = token
        self.expected = expected     # XXX str shouldn't necessary
        self.line = getattr(token, 'line', '?')
        self.column = getattr(token, 'column', '?')
        self.considered_rules = considered_rules
        self.state = state
        self.pos_in_stream = getattr(token, 'pos_in_stream', None)

        message = ("Unexpected token %r at line %s, column %s.\n"
                   "Expected one of: \n\t* %s\n"
                   % (token, self.line, self.column, '\n\t* '.join(self.expected)))

        super(UnexpectedToken, self).__init__(message)

class VisitError(LarkError):
    def __init__(self, tree, orig_exc):
        self.tree = tree
        self.orig_exc = orig_exc

        message = 'Error trying to process rule "%s":\n\n%s' % (tree.data, orig_exc)
        super(VisitError, self).__init__(message)
###}



# -*- coding: utf-8 -*-
"""
This class aligns natural language phrases into their most default forms and matches them
with niceStr parameters located in templates_dic.yaml. Returns an dict with selected template and
filled as many fields from the dic as possible.
"""
import re, os, imp

from gherkan.containers.NLPModelManager import NLPModelManager
from gherkan.utils.word_morpho import Finetuner
from pattern.en import parse, parsetree, lemma, tag
import majka, yaml, string
import pkg_resources, json
import gherkan.utils.gherkin_keywords as g
import gherkan.utils.constants as c
from gherkan.utils import logging_types
import logging

GENERAL_YAML_DIR = os.path.join(c.GHERKAN_ROOT_DIR, "utils", "en-cz_phrases.yaml")

class NLTempMatcher():
    def __init__(self, lang, nmm : NLPModelManager, feature: str):
        self.actor_uni = feature
        self.finetune = Finetuner(lang, nmm)
        self.NLPhrase = None
        self.lang = lang
        self.subj_of_interest = c.SUBJ_OF_INTEREST
        self.program_dic = self.load_program_dic()
        self.morph_wlt = self.finetune.morph_wlt

    def get_NLPhrase(self, phrase):
        phrase = self.remove_passive(phrase)
        negative = self.get_negative(phrase)
        force, unforce = self.get_force_unforce(phrase)
        state, phrase = self.get_state(phrase)
        value = self.get_value(phrase)
        clean_phrase = self.get_cleanphr(phrase, negative, force, unforce)
        if self.lang == "en":
          clean_phrase = self.lemmatize(clean_phrase)
        pn = self.get_pn(clean_phrase)
        if pn:
            clean_phrase = pn
            clean_phrase = self.make_spec_upper(clean_phrase)
        else:
            clean_phrase = self.lemmatize(clean_phrase)
            clean_phrase = self.make_spec_upper(clean_phrase)
            clean_phrase = self.finetune.strip_extra_spaces(clean_phrase)
        clean_phrase = self.finetune.strip_extra_spaces(clean_phrase)
        self.NLPhrase = {"tempToNL": clean_phrase,
                         "negate": negative,
                         "force": force,
                         "unforce": unforce,
                         "state": state,
                         "value" : value
                         }
        return self.NLPhrase

    def remove_passive(self, phrase):
        if self.lang == "en":
            if " ".join(["by", self.subj_of_interest]) in phrase.lower():  # handle passive verb form
                tree = parsetree(phrase, tokenize=True, tags=True,
                                 chunks=True, relations=True)
                for x in tree:  # now we select sentences with robot subjects or passive form
                    if " ".join(["by", self.subj_of_interest]) in x.string.lower():  # handle passive verb form
                        verbs, nouns = ([] for i in range(2))
                        verb_object = (x.string).split(" by ")[0]
                        subject = (x.string).split(" by ")[1]
                        for chunk in x.chunks:
                            if chunk.type == "VP":
                                for word in chunk.words:
                                    verbs.append(
                                        (lemma(word.string)))
                                if "be" in verbs:
                                    verbs.remove("be")
                                rest = verb_object.replace(chunk.string, "")
                        sent = subject + " " + ' '.join(verbs) + " " + rest
                        phrase = sent.translate(
                            str.maketrans('', '', string.punctuation))
        elif self.lang == "cs":
            pass

        return phrase

    def make_spec_upper(self, sent):
        """e.g. shuttle xy > shuttle XY"""
        if self.lang == "en":
            tree = parsetree(sent, tokenize=True, tags=True, chunks=True, relations=True)
            for sentence in tree:
                actors = sentence.nouns
            if actors:
                for actor in actors:
                    spec = re.findall(r'{}\s+([A-Za-z0-9]{})\b'.format(actor.string,"{1,3}"), sent)
                    if spec:
                        for n in spec:
                            for word, pos in tag(n):
                                if pos == "NN" or pos == "NNP" or pos == "RB":
                                  sent = str(re.sub(r'\b({})\b'.format(n), n.upper(), sent))
        elif self.lang == "cs":
           nouns = []
           for word in sent.split():
            word_info = self.finetune.get_txt_info(word)
            if (word_info[0]["pos"]) == "N" and (word_info[0]["pos_det"]) == "N":
                nouns.append(word)
            for noun in nouns:
                if len(noun) <= 3:
                    sent = re.sub(r'\b{}\b'.format(noun), noun.upper(), sent)
        return sent

    def get_pn(self, phrase):
        actor = re.findall(r'{}\s+([A-Za-z0-9]{})\b'.format(self.subj_of_interest,"{1,2}"),
                           phrase, re.IGNORECASE)
        if not actor:
            actor_ori = self.subj_of_interest
            if self.subj_of_interest in self.actor_uni and self.subj_of_interest in phrase:
                actor = self.actor_uni
            elif self.subj_of_interest not in self.actor_uni:
                # logging.warning("Feature does not contain {}, cannot resolve phrase {}".format(self.subj_of_interest, phrase))
                actor = self.actor_uni
        else:
            actor = " ".join([self.subj_of_interest, actor[0]])
            actor_ori = actor
        verb = self.finetune.find_verb(self.lang, phrase)
        if verb:
            verb_lemma = self.finetune.lemmatize(self.lang, verb[0])
            phrase = re.sub(verb[0], verb_lemma, phrase)
        action_lemma = re.sub(r'\b{}\b'.format(actor_ori), "", phrase, re.IGNORECASE)
        action_lemma = self.lemmatize(action_lemma)
        if not actor:
            return False
        if actor.lower() in self.program_dic.keys():
            action_lemma = "".join(action_lemma.lower().rsplit(actor.lower()))
            action_lemma = self.finetune.lemmatize_sentence(self.lang, action_lemma)
            for key in self.program_dic[actor.lower()]:
                if self.finetune.strip_extra_spaces(action_lemma.lower()) == self.program_dic[actor.lower()][key].lower():
                    actor_l = actor.split()
                    actor = " ".join([actor_l[0], actor_l[1].upper()])
                    phrase = "{} #{}".format(actor, key) # internal format for program numbers in NL
                    return phrase
        elif actor.lower() == self.subj_of_interest:  # if no specification, use the one defined in Feature
            for key in self.program_dic:
                action_lemma = self.finetune.lemmatize_sentence(self.lang, action_lemma)
                for value in self.program_dic[key]:
                 if self.finetune.strip_extra_spaces(action_lemma) == self.program_dic[key][value]:
                    phrase = "{} #{}".format(key, value) # internal format for program numbers in NL
                    return phrase
        return False

    def translate(self, phrase):
        new_phrase = phrase
        with open(GENERAL_YAML_DIR, 'r', encoding="utf-8") as stream:
            try:
                vocab = yaml.full_load(stream)
            except:
                logging.debug("Could not load en-cz_phrases {}", GENERAL_YAML_DIR)
        for word in phrase.split():
            for key in vocab["cs"]:
                word_en = ""
                if isinstance(vocab['cs'][key], list):
                   for x in vocab['cs'][key]:
                      if word.lower() == x:
                          word_en = key
                elif word.lower() == vocab['cs'][key]:
                  word_en = key
                if word_en:
                    new_phrase = re.sub(r'\b{}\b'.format(word), word_en, new_phrase, re.IGNORECASE)
        return new_phrase

    def get_state(self, phrase):
        # TODO infinitive should be enough if the phrases are lemmatized
        stop_phrases=["end", "ends", "ended", "ending", "finish", "finishes", "stops", "finished", "stopped",
                      "stop", "skončit", "dokončit", "přestat", "-skončit", "-dokončit", "end", "finish", "do start"]
        start_phrases=["start", "starts", "started", "starting", "začít"]
        for word in phrase.split():
                if self.finetune.lemmatize(self.lang, word).lower() in stop_phrases:
                   if word == phrase.split()[-1] or phrase.split()[-1] == "program":
                       break
                   phrase = re.sub(word, '', phrase, re.IGNORECASE)
                   state = "End"
                   return state, phrase
                elif self.finetune.lemmatize(self.lang, word).lower() in start_phrases:
                    if word == phrase.split()[-1]:
                        break
                    phrase = re.sub(word, '', phrase, re.IGNORECASE)
                    state = "Start"
                    return state, phrase
        state = None
        return state, phrase


    def get_value(self, phrase):
        # TODO only a stub
        # this method should discover value in the phrase which is stated only implicitly, i.e. which is not in the template

        value = None
        zero_value_phrases = ["empty", "prázdný"]

        for word in phrase.split():
            if self.finetune.lemmatize(self.lang, word).lower() in zero_value_phrases:
                value = 0

        return value


    def get_force_unforce(self, phrase):
        if self.lang == 'en':
           force = re.findall(r'\b(force)\b', phrase, re.IGNORECASE)
           unforce = re.findall(r'\b(unforce)\b', phrase, re.IGNORECASE)
        elif self.lang == 'cs':
            no_diacritics = self.finetune.remove_diacritics(phrase)
            force = re.findall(r'\b({}|{})\b'.format("Vynut", "Vynutit"), no_diacritics, re.IGNORECASE)
            unforce = re.findall(r'\b({})\b'.format("Uvolni", "Uvolnit"), no_diacritics, re.IGNORECASE)
        if not self.empty_tree(force):
            force = True
        else:
            force = False
        if not self.empty_tree(unforce):
            unforce = True
        else:
            unforce = False
        return force, unforce

    def get_negative(self, phrase):
        verbs = self.finetune.find_verb(self.lang, phrase)
        if self.lang == "en":
            if verbs:
                for verb in verbs:
                    negative = re.findall(r'\b(not)\b\s+\b({})\b|(\b({})\b)\s\b(not)\b'.format(verb, verb), phrase)
                    if not self.empty_tree(negative):
                        negative = True
                    else:
                        negative = False
            else:
                negative = False
        elif self.lang == "cs":
            negative = False
            verbs = self.finetune.find_verb(self.lang, phrase)
            if verbs:
                for verb in verbs:
                    verb_info = self.finetune.get_txt_info(verb)
                    if verb_info[0]["negate"] == "N":
                        negative = True
        return negative

    def get_cleanphr(self, phrase, negative=None, force=None, unforce=None):
        # clean the phrase from negatives and force
        del_words = ["do", "does", "will", "can", "must", "have to"]
        if self.lang == "en":
            if negative:
              phrase = re.sub(r'\b(not)\b', '', phrase, re.IGNORECASE)
              for i in del_words:
                phrase = re.sub(r'\b({})\b'.format(i), "", phrase)
                phrase = self.finetune.strip_extra_spaces(phrase)
            if force:
              phrase = re.sub(r'\b(force)\b', '', phrase, re.IGNORECASE)
            if unforce:
                phrase = re.sub(r'\b(unforce)\b', '', phrase, re.IGNORECASE)
            # remove excess verb words
        elif self.lang == "cs":
            if force:
              phrase = re.sub(r'\b({}|{}|{})\b'.format("vynuť", "vynut", "vynutit"), '', phrase, re.IGNORECASE)
            if unforce:
                phrase = re.sub(r'({})\b'.format("uvolni", "uvolnit"), '', phrase, re.IGNORECASE)
            if negative:
                verbs = self.finetune.find_verb(self.lang, phrase)
                if verbs:
                    for verb in verbs:
                        verb_info = self.finetune.get_txt_info(verb)
                        if verb_info[0]["negate"] == "N":
                            infinitive = self.finetune.lemmatize(self.lang, verb).split("-")
                            if infinitive:
                              verb_infinitiv = "".join(string for string in infinitive if len(string) > 0)
                              verb_new = re.sub(r"^(ne)", "", verb_infinitiv, re.IGNORECASE)
                              phrase = re.sub(verb, verb_new, phrase, re.IGNORECASE)
                            else:
                                verb_new = re.sub(r"^(ne)","", verb, re.IGNORECASE)
                                phrase = re.sub(verb, verb_new, phrase, re.IGNORECASE)

        phrase = self.finetune.strip_extra_spaces(phrase)
        return phrase

    def empty_tree(self, input_list):
        """Recursively iterate through values in nested lists."""
        for item in input_list:
            if not isinstance(item, list) or not self.empty_tree(item):
                return False
        return True

    def lemmatize(self, text):
        with pkg_resources.resource_stream("gherkan", "utils/verbs.yaml") as stream:
            verb_dic = yaml.full_load(stream)
        with pkg_resources.resource_stream("gherkan", "utils/lemma_exceptions.yaml") as stream:
            lemma_dic = yaml.full_load(stream)
        lemmas = ""
        skip_list = [",", ".", ";"]
        if self.lang == "cs":
            verb = self.finetune.find_verb(self.lang, text)
            if verb:
                for key in verb_dic[self.lang]["verbs"]:
                    if verb[0] in key.keys():
                        verb_lemma = key[verb[0]]
                        text = re.sub(r'\b{}\b'.format(verb[0]), verb_lemma, text, re.IGNORECASE)
            self.morph_wlt.flags |= majka.ADD_DIACRITICS  # find word forms with diacritics
            self.morph_wlt.tags = False  # return just the lemma, do not process the tags
            self.morph_wlt.first_only = True  # return only the first entry
            for word in text.split():
               if word in skip_list:
                     pass
               else:
                   found = False
                   for key in lemma_dic[self.lang]:
                       for keyword in lemma_dic[self.lang][key]:
                         if word.lower() == keyword:
                           lemmas += " "
                           lemmas += key
                           found = True
                           break
                   if not found:
                       output = self.morph_wlt.find(word)
                       lemmas += " "
                       if output:
                         lemma_det = output[0]['lemma']
                         lemmas += lemma_det
                       else:
                           lemmas += word

        elif self.lang == "en":
            # this version keeps capitals
            tree = parsetree(text, tokenize=True)

            for sentence in tree:
                for word in sentence:
                    if word.string in skip_list:
                        pass
                    found = False
                    for key in lemma_dic[self.lang]:
                        for keyword in lemma_dic[self.lang][key]:
                            if word.string.lower() == keyword:
                                lemmas += " "
                                lemmas += key
                                found = True
                                break
                    if not found:
                        lemmas += " "
                        lemmas += lemma(word.string)

        return lemmas

    def load_program_dic(self):
        # load program json
        program_dic = {}
        if self.lang == "en":
            file = "utils/RobotPrograms_en.json"
        elif self.lang == "cs":
            file = "utils/RobotPrograms_cs.json"
        with pkg_resources.resource_stream("gherkan", file) as stream:
            try:
                program_dic = json.load(stream)
            except:
                logging.debug("Could not load {}", file)
        return program_dic




import sys

from json_parser import Lark_StandAlone, Transformer, inline_args

class TreeToJson(Transformer):
    @inline_args
    def string(self, s):
        return s[1:-1].replace('\\"', '"')

    array = list
    pair = tuple
    object = dict
    number = inline_args(float)

    null = lambda self, _: None
    true = lambda self, _: True
    false = lambda self, _: False


parser = Lark_StandAlone(transformer=TreeToJson())

if __name__ == '__main__':
    with open(sys.argv[1]) as f:
        print(parser.parse(f.read()))




from gherkan.containers.NLPModelManager import NLPModelManager
from gherkan.decoder.NLParser import NLParser
from gherkan.encoder.RawNLParser import RawNLParser
from gherkan.encoder.SignalFileWriter import SignalFileWriter

import logging

def nl_to_signal(base_path: str, request: dict, nmm: NLPModelManager = None, splitter=None):
    """
    Processes raw text and saves two files - intermediate NL file (base_path.feature) and
    resulting signal file (base_path_signals.feature).

    Parameters
    ----------
    base_path : str
        path/to/files/file_basename

    request : dict
        dictionary with raw text
    """

    # create paths to files based on base_path
    nl_file_path = base_path + ".feature"
    signal_file_path = base_path + "_signals.feature"

    if nmm is None:
        logging.info("Loading all NLP models")
        nmm = NLPModelManager()
        nmm.load_all_models()

    language = request["language"]

    if splitter is None:
        splitter = "scénář" if language == "cs" else "scenario"

    # parse the raw text
    parser = RawNLParser(nmm=nmm, feature=request["feature"], language=language, splitter=splitter)
    parser.parse(request["text_raw"])
    request["background"] = parser.parse(request["background"], background=True)
    ## parser.generate_program_dict(make_new=False) to extend the json with robot programs
    ## parser.generate_program_dict() to always make new json
    parser.generate_program_dict(make_new=False)
    raw_text_lines = parser.get_text_lines()

    # create intermediate representation - text lines of the NL file
    lines = [
        "# language: {}".format(request["language"]),
        "Feature: {}".format(request["feature"]),
        "  {}".format(request["feature_desc"]),
        "Background:",
        "  {}".format(request["background"])
    ]
    lines += raw_text_lines

    # save the intermediate NL file
    with open(nl_file_path, "wt", encoding="utf-8") as out:
        out.write("\n".join(lines))

    print("Saved NL file as {}".format(nl_file_path))

    # parse the NL text lines
    nl_parser = NLParser()
    nl_batch = nl_parser.parse(lines)

    # write the signal file
    signalFileWriter = SignalFileWriter(nmm=nmm)
    signalFileWriter.encode(nl_batch)
    signalFileWriter.write(signal_file_path)

    print("Saved signals file as {}".format(signal_file_path))



from gherkan.containers.StatementTreeNode import StatementTreeNode, StatementTreeBinaryOperatorNode
from gherkan.decoder.SignalParser import SignalParser


def processNode(node: StatementTreeNode):
    """This function will process a tree node and return a string depending on the node parameters

    Arguments:
        node {SignalStatementTreeNode} -- The tree node

    Returns:
        str -- A string representing the tree node
    """

    # Obviously, the return value will depend on what we want to do with the node
    return node.kind + "; "


def recurseIntoTree(node: StatementTreeNode):
    """This function will recursively traverse the tree deeper and deeper
       until all the nodes are processed with the "processNode" function.

    Arguments:
        node {SignalStatementTreeNode} -- The tree node

    Returns:
        str -- A result string. This can be anything depending on what the process function should return.
    """

    # Apply the function onto this node
    result = processNode(node)

    # If the current node type is "operator", i.e. it has children,
    # traverse deeper into its children
    if type(node) is StatementTreeBinaryOperatorNode:
        # Retrieve result from the left child
        result += recurseIntoTree(node.lchild)
        # and then retrieve result from the right child
        result += recurseIntoTree(node.rchild)

    # Return the result
    return result


if __name__ == "__main__":
    # Parse statement into the SignalStatementTree
    signalParser = SignalParser()
    st = signalParser.parseStatement("part2OnTable == 1 && (robotN1ProgramNumber == 0 || robotN1ProgramNumber == 1 || robotN1ProgramEnded) && edge(robotN3ProgramEnded, 1)")

    # Call the recursive function with the tree's *root* as an argument
    result = recurseIntoTree(st.root)

    # print result
    print(result)



# -*- coding: utf-8 -*-
"""
"""
from gherkan.containers.StatementTreeNode import StatementTreeBinaryOperatorNode, \
    StatementTreeNode, StatementTreeOperandNode, StatementTreeMergedNode
from gherkan.containers.StatementTree import StatementTree
from gherkan.decoder.SignalParser import SignalParser

import gherkan.utils.constants as c
import gherkan.utils.gherkin_keywords as g
from gherkan.utils import logging_types
import json

import logging


class SignalMapper():

    def __init__(self):
        self.batch = None
        self.language = None
        self.feature = ""
        self.dictionary = {}

    def loadSignal(self, signalFilePath: str):
        sp = SignalParser()
        self.outLines = []
        self.batch = sp.parseFile(signalFilePath)
        self.language = self.batch.language
        self.feature = self.batch.name

    # <=== Analysis ===>
    def analyze(self):
        if self.batch is None:
            Exception("Signal file was not loaded, yet!")
        batch = self.batch

        if batch.context:
            self.analyze_node(batch.context.root)

        for scenario in batch.scenarios:
            for tree in scenario.givenStatements:
                self.analyze_node(tree.root)
            for tree in scenario.whenStatements:
                self.analyze_node(tree.root)
            for tree in scenario.thenStatements:
                self.analyze_node(tree.root)

    def analyze_node(self, node: StatementTreeNode):
        if isinstance(node, StatementTreeBinaryOperatorNode):
            self.analyze_operator(node)
        elif isinstance(node, StatementTreeMergedNode):
            self.analyze_merged_operator(node)
        elif isinstance(node, StatementTreeOperandNode):
            self.analyze_operand(node)

    def analyze_merged_operator(self, node: StatementTreeMergedNode):
        for subnode in node.subnodes:
            self.analyze_operand(subnode)

    def analyze_operator(self, node: StatementTreeBinaryOperatorNode):
        if not (node.kind == c.AND or node.kind == c.OR):
            raise NotImplementedError("Unrecognized operator {}".format(node.kind))
        self.analyze_node(node.lchild)
        self.analyze_node(node.rchild)

    def analyze_operand(self, node: StatementTreeOperandNode):
        signal = node.data.variable
        if signal not in self.dictionary:
            self.dictionary[signal] = signal

    def writeDictionary(self, dictFilePath: str):
        with open(dictFilePath, "w", encoding="utf-8") as dictFile:
            json.dump(self.dictionary, dictFile, ensure_ascii=False, indent=4)

    def loadDictionary(self, dictFilePath: str):
        with open(dictFilePath, "r", encoding="utf-8") as dictFile:
            self.dictionary = json.load(dictFile)

    # <=== Signal Writing ===>
    def encode(self):
        if self.batch is None:
            Exception("Signal file was not loaded, yet!")
        batch = self.batch

        self.outLines.append("# language: {}\n\n".format(self.language))
        self.outLines.append("{}: {}\n".format(g.get_kw(self.language, "Feature"), batch.name))
        self.outLines.append("  {}\n".format(batch.desc))

        if batch.context:
            self.outLines.append("{}:\n".format(g.get_kw(self.language, "Background")))
            self.outLines.append("  {} {}\n\n".format(g.get_kw(self.language, "Given"), self.tree_to_str(batch.context)))

        for scenario in batch.scenarios:
            self.outLines.append("{}: {}\n".format(g.get_kw(self.language, "Scenario"), scenario.name))

            for tree in scenario.givenStatements:
                self.outLines.append("  {} {}\n".format(g.get_kw(self.language, "Given"), self.tree_to_str(tree)))
            for tree in scenario.whenStatements:
                self.outLines.append("  {} {}\n".format(g.get_kw(self.language, "When"), self.tree_to_str(tree)))
            for tree in scenario.thenStatements:
                self.outLines.append("  {} {}\n".format(g.get_kw(self.language, "Then"), self.tree_to_str(tree)))

    def writeSignal(self, outputFilePath: str):
        with open(outputFilePath, "w", encoding="utf-8") as out:
            out.writelines(self.outLines)

    def tree_to_str(self, tree: StatementTree):
        return self.node_to_str(tree.root)

    def node_to_str(self, node: StatementTreeNode):
        if type(node) == StatementTreeBinaryOperatorNode:
            return self.operator_to_str(node)
        elif type(node) == StatementTreeMergedNode:
            return self.merged_operator_to_str(node)
        elif type(node) == StatementTreeOperandNode:
            return self.operand_to_str(node)

    def merged_operator_to_str(self, node: StatementTreeMergedNode):
        return " && ".join([self.operand_to_str(subnode) for subnode in node.subnodes])

    def operator_to_str(self, node: StatementTreeBinaryOperatorNode):
        if node.kind == c.AND:
            operator = "&&"
        elif node.kind == c.OR:
            operator = "||"
        else:
            raise NotImplementedError("Unrecognized operator {}".format(node.kind))

        return ("({}) {} ({})".format(
            self.node_to_str(node.lchild),
            operator,
            self.node_to_str(node.rchild)))

    def operand_to_str(self, node: StatementTreeOperandNode):
        kind = node.kind
        variable = node.data.variable
        if variable in self.dictionary:
            variable = self.dictionary[variable]
        else:
            logging.warning("Signal name not found in the mapping dictionary: {}; keeping the original name.", variable, extra={"type": logging_types.W_MAPPING_NOT_DEFINED, "phrase": str(node)})

        outString = ""
        if kind == c.EQUALITY:
            outString = "{} {}= {}".format(
                variable,
                "!" if node.negated else "=",
                node.data.value
            )
        elif kind == c.BOOL:
            outString = "{}{}".format(
                # XOR in case boolean is negated by the 'negated' attribute
                "" if (node.data.value ^ node.negated) else "! ",
                variable,
            )
        elif kind in [c.EDGE,
                      c.FORCE,
                      c.UNFORCE]:
            outString = "{}{}({}, {})".format(
                "! " if node.negated else "",
                node.kind.lower(),
                variable,
                node.data.value
            )
        else:
            logging.warning("Node kind {} not recognized!", kind, extra={"type": logging_types.W_OPERAND_UNKNOWN, "phrase": str(node)})

        return outString


# signalFileMapper = SignalMapper()
# signalFileMapper.loadSignal(r"C:\Users\admin\Documents\nl_instruction_processing\data\input\Scenare\Montrac_signals.feature")
# # signalFileMapper.analyze()

# # signalFileMapper.dictionary["robotR3ProgamEnd"] = "programEnd"
# signalFileMapper.loadDictionary(r"C:\Users\admin\Documents\test.json")
# signalFileMapper.encode()



import requests
import unittest
import os
import gherkan.utils.gherkin_keywords as g
import gherkan.utils.constants as c



class TestAudio(unittest.TestCase):

    def setUp(self):
        self.url = f"http://{self.__class__.host}:{self.__class__.port}/{{}}"

        resetResponse = requests.get(self.url.format("reset"))
        self.assertDictContainsSubset({"OK": True}, resetResponse.json())


    def compare_multiline(self, testee, teststr):
        if not isinstance(testee, str) or not isinstance(teststr, str):
            self.assertFalse("No strings given to compare!")
            return

        self.assertEqual(testee.lower().replace("\n", " ").split(),
                         teststr.lower().replace("\n", " ").split())


    def transcribeAudio(self, lang, audioFilename):
        request = {
            "language": lang,
        }
        files = {'upload_file': open(audioFilename, 'rb')}
        response = requests.post(self.url.format("audio"), data=request, files=files)
        try:
            response = response.json()
        except Exception as exception:
            self.fail(
                "An exception was raised while trying to convert response to JSON. "
                "This is most likely due to the fact that the response was of different type.\n"
                "The exception: {}\nResponse text (if possible):\n{}".format(
                    exception, response.text))

        if not "transcript" in response:
            raise Exception(response["message"])

        return response["transcript"]


    def rawTextToSignal(self, lang, transcript):
        request = {
            "feature": "Montrac",
            "feature_desc": "Test audia",
            "background": "Jakmile linka je zapnutá",
            "text_raw": transcript,
            "language": lang
        }
        response = requests.post(self.url.format("nltext"), data=request)
        try:
            response = response.json()
        except Exception as exception:
            self.fail(
                "An exception was raised while trying to convert response to JSON. "
                "This is most likely due to the fact that the response was of different type.\n"
                "The exception: {}\nResponse text (if possible):\n{}".format(
                    exception, response.text))
        else:
            self.assertDictContainsSubset({"errors": False}, response)

            response = requests.get(self.url.format("signals"))
            return response


    def singleAudioTest(self, lang, audioFilename, correctResponse):
        transcript = self.transcribeAudio(lang, audioFilename)
        response = self.rawTextToSignal(lang, transcript)

        self.compare_multiline(response.content.decode(), correctResponse)



    # =============== TESTS ===============
    def test_audio(self):
        lang="cs"
        audioFilename=os.path.join(os.path.dirname(os.path.realpath(__file__)),os.pardir,"speech_recognition","montrac.wav")

        correctResponse = """# language: cs
            Požadavek: Montrac
                Test audia
            Kontext:
              Pokud lineOn
            
            Scénář: jakmile stanice z je volná pak vozíček y jede do stanice z
              Když stationzFree
              Pak shuttleYDestination == z
            Scénář: jakmile robot R1 zvedne kostku 1 pak robot R1 položí kostičku 1 na vozíček y na pozici 1
              Když robotR1ProgramNumber == 2
              Pak robotR1ProgramNumber == 10
            Scénář: když stanice x je volná a robot R1 skončil vykládání kostičky 1 pak vozíček y jede do stanice X
              Když (stationxFree) && (robotR1ProgramNumber == 11 && robotR1ProgramEnded)
              Pak shuttleYDestination == X
            """

        self.singleAudioTest(lang, audioFilename, correctResponse)


if __name__ == '__main__':
    unittest.TestCase.host = "localhost"
    unittest.TestCase.port = 5000
    unittest.main()



# coding: utf-8

import yaml
import re
from pattern.en import conjugate
import imp
import os

CZ_PROGRAM_YAML_DIR = os.path.join(imp.find_module("gherkan")[1], "utils", "ProgramMeanings_cz.yaml")
EN_PROGRAM_YAML_DIR = os.path.join(imp.find_module("gherkan")[1], "utils", "ProgramMeanings_en.yaml")
GENERAL_YAML_DIR = os.path.join(imp.find_module("gherkan")[1], "utils", "General_phrases.yaml")
SCENARIO = "ProgramNumbers_Scene_2"


class FinetuneLang:

    def conjugate_en(self, word, person, number="singular"):
        if number == "singular":
            self.tag = str(person) + "sg"
            self.declenation = conjugate(word, self.tag)

        elif number == "plural":
            self.tag = str(person) + "pl"
            self.declenation = conjugate(word, self.tag)

        return self.declenation


class FillActions:
    def __init__(self):
        self.finetune = FinetuneLang()

    def ParseMainAction(self, data):
        self.data = data
        if "robot" in self.data[0][0] and "Program" in self.data[3]:
            self.rob_name = "Robot " + self.data[0][0].split('robot')[1]
            if len(self.data[0]) == 2:
                self.rob = self.data[0][0]
                self.prog = self.data[0][1]
                self.action_en = self.MatchActionEN(self.rob, self.prog)
                self.action_cz = self.MatchActionCZ(self.rob, self.prog)
                if "equality" in self.data[1]:
                    self.action_cz = self.rob_name + " " + self.action_cz
                    self.declen_en = self.finetune.conjugate_en(self.action_en.split()[0], 3)
                    self.action_en = self.rob_name + " " + self.action_en.replace(self.action_en.split()[0], self.declen_en)
                elif "equality" not in self.data[1]:
                    self.action_en, self.action_cz = self.ParseBool_Robot(self.data, [self.action_en, self.action_cz])
                    self.action_cz = self.rob_name + " " + self.action_cz
                    self.action_en = self.rob_name + " " + self.action_en
            elif len(self.data[0]) == 1:
                self.action_en, self.action_cz = self.ParseBool_Robot(self.data)
                self.action_cz = self.rob_name + " " + self.action_cz[:-1]
                self.action_en = self.rob_name + " " + self.action_en[:-1]

        elif "AtStation" in self.data[3] or "Destination" in self.data[3]:
            self.actor = re.findall('[a-zA-Z][^A-Z]*', self.data[0][0])
            self.actor_name = self.actor[0] + " " + ''.join(self.actor[1:])
            if "AtStation" in self.data[3]:
                self.matched_en = self.MatchGeneral("AtStation", lang="Phrases_En", key=int(self.data[2]))
                self.matched_cz = self.MatchGeneral("AtStation", lang="Phrases_Cz", key=int(self.data[2]))
            else:
                self.matched_en = self.MatchGeneral("Destination", lang="Phrases_En", key=int(self.data[2]))
                self.matched_cz = self.MatchGeneral("Destination", lang="Phrases_Cz", key=int(self.data[2]))
            self.declen_en = self.finetune.conjugate_en(self.matched_en.split()[0], 3)
            self.action_en = self.actor_name + " " + self.matched_en.replace(self.matched_en.split()[0], self.declen_en) + " " + self.data[0][1]
            self.action_cz = self.MatchGeneral(self.actor[0], lang="Phrases_Cz") + " " + ''.join(self.actor[1:]) + " " + self.matched_cz + " " + self.data[0][1]

        elif "Free" in self.data[3]:
            self.actor = re.findall('[a-zA-Z][^A-Z]*', self.data[0][0])
            self.actor_name = self.actor[0] + " " + ''.join(self.actor[1:])
            self.matched_en = self.MatchGeneral("Free", lang="Phrases_En", key=int(self.data[2]))
            self.matched_cz = self.MatchGeneral("Free", lang="Phrases_Cz", key=int(self.data[2]))
            self.declen_en = self.finetune.conjugate_en(self.matched_en.split()[0], 3)
            self.action_en = self.actor_name + " " + self.matched_en.replace(self.matched_en.split()[0], self.declen_en)
            self.action_cz = self.MatchGeneral(self.actor[0], lang="Phrases_Cz") + " " + ''.join(self.actor[1:]) + " " + self.matched_cz

        elif "Move" in self.data[3]:
            self.actor = re.findall('[a-zA-Z][^A-Z]*', self.data[0][0])
            self.actor_name = self.actor[0] + " " + ''.join(self.actor[1:])
            self.matched_en = self.MatchGeneral("Move", lang="Phrases_En", key=int(self.data[2]))
            self.matched_cz = self.MatchGeneral("Move", lang="Phrases_Cz", key=int(self.data[2]))
            self.declen_en = self.finetune.conjugate_en(self.matched_en.split()[0], 3)
            self.action_en = self.actor_name + " " + self.matched_en.replace(self.matched_en.split()[0], self.declen_en)
            self.action_cz = self.MatchGeneral(self.actor[0], lang="Phrases_Cz") + " " + ''.join(self.actor[1:]) + " " + self.matched_cz

        elif "Full" in self.data[3]:
            self.actor = re.findall('[a-zA-Z][^A-Z]*', self.data[0][0])
            self.actor_name = self.actor[0] + " " + ''.join(self.actor[1:])
            self.matched_en = self.MatchGeneral("Full", lang="Phrases_En", key=int(self.data[2]))
            self.matched_cz = self.MatchGeneral("Full", lang="Phrases_Cz", key=int(self.data[2]))
            self.declen_en = self.finetune.conjugate_en(self.matched_en.split()[0], 3)
            self.action_en = self.actor_name + " " + self.matched_en.replace(self.matched_en.split()[0], self.declen_en)
            self.action_cz = self.MatchGeneral(self.actor[0], lang="Phrases_Cz") + " " + ''.join(self.actor[1:]) + " " + self.matched_cz

        elif "Count" in self.data[3]:
            self.actor = re.findall('[a-zA-Z][^A-Z]*', self.data[0][0])
            self.actor_name = self.actor[0] + " " + ''.join(self.actor[1:])
            if self.data[2] == "0":
                self.action_en = self.actor_name + " is empty"
                self.action_cz = self.MatchGeneral(self.actor[0], lang="Phrases_Cz") + " je prázdný"
            else:
                self.action_en = "Number of objects in " + self.actor_name + " is " + str(self.data[2])
                self.action_cz = "Počet objektů ve " + self.MatchGeneral(self.actor[0], lang="Phrases_Cz", key=2) + " " + ''.join(self.actor[1:]) + " " + " je " + str(self.data[2])

        elif "Locked" in self.data[3]:
            self.actor = re.findall('[a-zA-Z][^A-Z]*', self.data[0][0])
            self.actor_name = self.actor[0] + " " + ''.join(self.actor[1:])
            self.matched_en = self.MatchGeneral("Locked", lang="Phrases_En", key=int(self.data[2]))
            self.matched_cz = self.MatchGeneral("Locked", lang="Phrases_Cz", key=int(self.data[2]))
            self.declen_en = self.finetune.conjugate_en(self.matched_en.split()[0], 3)
            self.action_en = self.actor_name + " " + self.matched_en.replace(self.matched_en.split()[0]+" ", self.declen_en + " ")
            self.action_cz = self.MatchGeneral(self.actor[0], lang="Phrases_Cz") + " " + ''.join(self.actor[1:]) + " " + self.matched_cz

        elif "Replenish" in self.data[3]:
            self.actor = re.findall('[a-zA-Z][^A-Z]*', self.data[0][0])
            self.actor_name = self.actor[0] + " " + ''.join(self.actor[1:])
            self.matched_en = self.MatchGeneral("Replenish", lang="Phrases_En", key=int(self.data[2]))
            self.matched_cz = self.MatchGeneral("Replenish", lang="Phrases_Cz", key=int(self.data[2]))
            self.declen_en = self.finetune.conjugate_en(self.matched_en.split()[0], 3)
            self.action_en = self.actor_name + " " + self.matched_en.replace(self.matched_en.split()[0], self.declen_en)
            self.action_cz = self.MatchGeneral(self.actor[0], lang="Phrases_Cz") + " " + ''.join(self.actor[1:]) + " " + self.matched_cz

        elif "to" in self.data[3]:
            # TODO: Raise a warning?
            pass

        return [self.action_en, self.action_cz]

    def ParseBool_Robot(self, data, phrases=None):
        if phrases:
            self.eng_phrase = phrases[0]
            self.czech_phrase = phrases[1]
            self.data = data
        else:
            self.eng_phrase = ""
            self.czech_phrase = ""
            self.data = data

        if "Started" in self.data[3]:
            if self.data[2] == "1":
                self.czech_phrase_spec = ' '.join(["začne program", self.czech_phrase])
                self.eng_phrase_spec = ' '.join(["starts the program", self.eng_phrase])
            elif self.data[2] == "0":
                self.czech_phrase_spec = ' '.join(["nezačne program", self.czech_phrase])
                self.eng_phrase_spec = ' '.join(["does not start the program", self.eng_phrase])

        elif "Ended" in self.data[3]:
            if self.data[2] == "1":
                self.czech_phrase_spec = ' '.join(["skončí program", self.czech_phrase])
                self.eng_phrase_spec = ' '.join(["finishes program", self.eng_phrase])
            elif self.data[2] == "0":
                self.czech_phrase_spec = ' '.join(["neskončí program", self.czech_phrase])
                self.eng_phrase_spec = ' '.join(["does not finish the program", self.eng_phrase])

        return self.eng_phrase_spec, self.czech_phrase_spec

    def ParseNegative(self, phrase):
        self.eng_phrase = phrase[0]
        self.czech_phrase = phrase[1]

        self.czech_phrase.replace("skončí", "neskončí")
        self.czech_phrase.replace("začne", "nezačne")
        self.eng_phrase.replace("finishes", "does not finish")
        self.eng_phrase.replace("starts", "does not start")

        return self.eng_phrase, self. czech_phrase

    def MatchActionEN(self, actor, prog):
        with open(EN_PROGRAM_YAML_DIR, 'r', encoding="utf-8") as stream:
            self.action_list = yaml.load(stream)
            self.action_en = self.action_list[SCENARIO][actor][int(prog)]

        return self.action_en

    def MatchActionCZ(self, actor, prog):
        with open(CZ_PROGRAM_YAML_DIR, 'r', encoding="utf-8") as stream:
            self.action_list = yaml.load(stream)
            self.action_cz = self.action_list[SCENARIO][actor][int(prog)]

        return self.action_cz

    def MatchGeneral(self, data, lang, key=1):
        with open(GENERAL_YAML_DIR, 'r', encoding="utf-8") as stream:
            self.phrase_list = yaml.load(stream)
            self.phrase = self.phrase_list[lang][data][key]

        return self.phrase


if __name__ == "__main__":
    fill_action = FillActions()
    VarsVals = ['robotR1', '4']
    Type = "equality"
    Values = "4"
    PhraseStr = "ProgramNumber"
    TempData = [VarsVals, Type, Values, PhraseStr]
    print(TempData)
    sentence_core_part = fill_action.ParseMainAction(TempData)
    # sentence_core_spec = fill_action.ParseBool(sentence_core_part, SourcePhrase["TypeArr"][0])
    print(sentence_core_part)


import unittest
import os
import sys
"""
This code will gather and run all the tests in this directory. Files with tests must start with the "test_" prefix.
"""

host = "localhost"
port = 5000
if len(sys.argv) >= 2:
    host = sys.argv[1]
if len(sys.argv) == 3:
    port = int(sys.argv[2])

unittest.TestCase.host = str(host)
unittest.TestCase.port = int(port)

loader = unittest.TestLoader()
testSuite = loader.discover(os.path.dirname(os.path.abspath(__file__)))

testRunner = unittest.TextTestRunner(verbosity=2)
testRunner.run(testSuite)



# -*- coding: utf-8 -*-
import re
import logging
from gherkan.containers.StatementTreeNode import StatementTreeNode, StatementTreeOperandNode, StatementTreeBinaryOperatorNode
from lark.tree import Tree
from gherkan.utils import constants as c
from lark import Token
from gherkan.utils import logging_types


class StatementTree():
    def __init__(self, statement : str):
        self.binaryOperators = [c.AND, c.OR]
        self.equalities = [c.EQUALITY, c.INEQUALITY]
        self.statement = statement
        self.root = None

    def buildFromSignalTree(self, tree, negate=False):
        self.root = self.buildSignalNode(tree, parent=None, negate=negate)

    def buildFromNLTree(self, tree):
        self.root = self.buildNLNode(tree)

    def buildSignalNode(self, larkNode, parent=None, negate=False):
        kind = larkNode.data.upper()  # node alias had to be lowercase in the grammar
        str = self.statement[larkNode.meta.start_pos:larkNode.meta.end_pos]
        node = None

        if kind in self.binaryOperators:
            node = StatementTreeBinaryOperatorNode(kind, parent=parent)
            node.lchild = self.buildSignalNode(larkNode.children[0], parent=node, negate=negate)
            node.rchild = self.buildSignalNode(larkNode.children[1], parent=node, negate=negate)

        elif kind in self.equalities:
            node = StatementTreeOperandNode(kind, parent=parent)

            if kind == c.INEQUALITY:
                # there is not "inequality" as a node kind
                node.kind = c.EQUALITY
                negate = True

            node.data.string = str
            node.data.variable = larkNode.children[0].value
            node.data.value = larkNode.children[1].value

        elif kind == c.BOOL:
            node = StatementTreeOperandNode(kind, parent=parent)
            node.data.string = str
            node.data.variable = larkNode.children[0].value
            node.data.value = True

        elif kind == c.NOT:
            # do not save the node, just swap the negate flag and pass it on
            return self.buildSignalNode(larkNode.children[0], parent=parent, negate=not negate)

        elif kind == "FUNCTION":
            # edge, force etc. parsed by lark uniformly as "function"
            function_name = larkNode.children[0].data
            kind = function_name.upper()

            node = StatementTreeOperandNode(kind)
            node.data.string = function_name + str
            node.data.variable = larkNode.children[1].value
            node.data.value = larkNode.children[2].value

        else:
            logging.warning("Unrecognized node: {}", kind, extra={"type": logging_types.W_OPERAND_UNKNOWN, "phrase": str})

        if negate:
            node.negate()

        return node


    def buildNLNode(self, larkNode, parent=None, negate=False):
        if type(larkNode) == Token:
            node = StatementTreeOperandNode(kind=None)
            node.data.variableNL = larkNode.value.strip()
            return node

        kind = larkNode.data.upper()  # node alias had to be lowercase in the grammar
        node = None

        if kind in self.binaryOperators:
            node = StatementTreeBinaryOperatorNode(kind, parent=parent)
            node.lchild = self.buildNLNode(larkNode.children[0], parent=node, negate=negate)
            node.rchild = self.buildNLNode(larkNode.children[1], parent=node, negate=negate)
        elif kind == c.NOT:
            # do not save the node, just swap the negate flag and pass it on
            return self.buildNLNode(larkNode.children[0], parent=parent, negate=not negate)
        else:
            # TODO unrecognized node
            logging.warning("Unrecognized node: {}", kind, extra={"type": logging_types.W_OPERAND_UNKNOWN, "phrase": ""})

        if negate:
            node.negate()

        return node


    # TODO simplify
    def __printNode(self, node: StatementTreeNode, leftChild, levelsOpened):
        string = ''.join([AsciiArt.line + "\t" if lo else "\t" for lo in levelsOpened[:-1]])
        if leftChild:
            if node:
                string += AsciiArt.left + str(str(node) if bool(node) else "ERROR!!!") + '\n'
        else:
            if node:
                string += AsciiArt.right + str(str(node) if bool(node) else "ERROR!!!") + '\n'

        if type(node) is StatementTreeBinaryOperatorNode:
            string += self.__printNode(node.lchild, True, levelsOpened + [True])
            string += self.__printNode(node.rchild, False, levelsOpened + [False])
        return string

    def __str__(self):
        levelsOpened = [False]

        node = self.root
        string = str(node) + '\n'
        if type(node) is StatementTreeBinaryOperatorNode:
            string += self.__printNode(node.lchild, True, levelsOpened + [True])
            string += self.__printNode(node.rchild, False, levelsOpened + [False])
        return string


class AsciiArt:
    dash = '\u2500'
    line = '\u2502'
    left = '\u251c\u2500\u2500 '
    right = '\u2514\u2500\u2500 '

